// CoypExif.cpp : Defines the entry point for the console application.
//
//  Version 0.8
//      Ken Reneris     01/02/2005
//
//

#include "stdafx.h"
#ifdef _DEBUG
#define new DEBUG_NEW
#endif

using namespace std;

//
// Globals
//

CString     g_SrcDir;               // Main source tree
CString     g_DstDir;               // Main dest tree
BOOL        g_Recurs = FALSE;       // Recursively walk SrcDir
BOOL        g_Update = FALSE;       // Update non-matching tags as well?
BOOL        g_Replace = FALSE;      // Replace Exif info from Src into Dst
BOOL        g_DontDelete = FALSE;   // Don't delete original dst files
BOOL        g_DontUpdate = FALSE;   // Don't update dst files
BOOL        g_RemoveThumb = FALSE;  
UINTN       g_Verbose = 0;

UINTN       g_FileCount;
UINTN       g_CheckCount;
UINTN       g_UpdateCount;

UCHAR       *g_Buffer;
#define     G_BUFFER_SIZE   (1024*1024)

#define     VERBOSE_DIFF        0x01
#define     VERBOSE_SAVE        0x02
#define     VERBOSE_PARSE       0x04


EXIF_FORMAT g_ExifFormat[] = {
    1,  NULL,       // 0. undefined
    1,  NULL,       // 1. unsigned byte
    1,  NULL,       // 2. ascii string
    2,  Swiz16,     // 3. unsigned short
    4,  Swiz32,     // 4. unsigned long
    8,  SwizR64,    // 5. unsigned rational
    1,  NULL,       // 6. signed byte
    1,  NULL,       // 7. undefined
    2,  Swiz16,     // 8. signed short
    4,  Swiz32,     // 9. signed long
    8,  SwizR64,    // 10. signed rational
    4,  NULL,       // 11. signed float (not used)
    8,  NULL,       // 12. signed double (not used)
} ;


//
// Prototypes
//

VOID WalkDir(CString Dir);
VOID ProcessFile(CString SrcFile, CString DstFile);


// The one and only application object
CWinApp theApp;

void Usage()
{
    printf("usage: CopyExif [-ruox] [-v1..f] Src Dst\n");
    printf("\n");
    printf("    -r   Recursively walk Src folder\n");
    printf("    -u   Copy tags that don't match (else, just copy missing tags)\n");
    printf("    -s   Don't delete backup of original Dst files\n");
    printf("    -x   Don't update Dst files - show diffs only\n");
    printf("    -o   Overwrite Exif info in Dst (removes entries)\n");
    printf("    -t   Remove embedded thumbnail\n");
    printf("    -vn  Verbose output mask\n");
    printf("    -?   This page\n");
    printf("\n");
    printf("    Src  Source image file or folder\n");
    printf("    Dst  Destination image file or folder\n");
    printf("\n");
    printf("Copies/Adds Exif datum (from IFD0 & SubIFD) from Source to Destionation.\n");
    printf("If Source & Destionation are folders, the Exif copy is performed on\n");
    printf("all image files that appear in both the Source & Desitation folders.\n");
    printf("\n");
    printf("Note not all jpg exif formats are understood - in particular files\n");
    printf("encoded with big-endian format exif data will be skipped\n");
    printf("\n");
    printf("Version 0.8 - Ken Reneris\n");
    exit(1);
}


VOID ProcessFlags (TCHAR *Flag)
{
    for (Flag++; *Flag; Flag++) {
        switch (*Flag) {
        case 'r':
            g_Recurs = TRUE;
            break;

        case 'v':
            g_Verbose = xtoi(Flag+1);
            if (!g_Verbose) {
                Usage();
            }
            Flag += 1;
            break;

        case 'u':
            g_Update = TRUE;
            break;

        case 's':
            g_DontDelete = TRUE;
            break;

        case 'o':
            g_Replace = TRUE;
            break;

        case 'x':
            g_DontUpdate = TRUE;
            g_Verbose |= VERBOSE_DIFF;
            break;

        case 't':
            g_RemoveThumb = TRUE;
            break;

        case '?':
            Usage();
            break;

        default:
            printf("Unknown flag: %c\n", *Flag);
            Usage();
        }
    } 
}

int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])
{
    INTN    index;
    TCHAR   *arg;

	// initialize MFC and print and error on failure
	if (!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0))
	{
		// TODO: change error code to suit your needs
		_tprintf(_T("Fatal Error: MFC initialization failed\n"));
		return 1;
	}

    //
    // Crack args
    //

    for (index=1; index < argc; index++) {
        arg = argv[index];

        if (arg[0] == '-') {
            ProcessFlags(arg);

        } else if (!g_SrcDir.GetLength()) {
            g_SrcDir = arg;

        } else if (!g_DstDir.GetLength()) {
            g_DstDir = arg;

        } else {
            Usage();
        }
    }

    //
    //
    //

    g_Buffer = (UCHAR *) malloc(G_BUFFER_SIZE);
    ASSERT(g_Buffer);


    //
    // Sanity
    //

    if (g_SrcDir.GetLength() == 0 || g_DstDir.GetLength() == 0) {
        Usage();
    }

    g_SrcDir.MakeLower();
    g_DstDir.MakeLower();

    if (IsDir(g_SrcDir) && IsDir(g_DstDir)) {
        WalkDir(g_SrcDir);
        printf ("Files found %d, checked %d, updated %d\n", g_FileCount, g_CheckCount, g_UpdateCount);

    } else if (IsFile(g_SrcDir) && IsFile(g_DstDir)) {

        ProcessFile(g_SrcDir, g_DstDir);

    } else {

        printf("Src and Dst must either be directories or folders\n");
        Usage();
    }

    return 0;
}



VOID WalkDir(CString Dir)
{
    CString             FileName, Ext;
    HANDLE              h;
    WIN32_FIND_DATA     data;
    INTN                pos;

    //
    // Process the local files
    //

    FileName = Dir + "\\*";
    h = FindFirstFile(FileName, &data);
    if (h != INVALID_HANDLE_VALUE) {
        do {
            if (!(data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
                FileName = data.cFileName;
                FileName.MakeLower();
                pos = FileName.ReverseFind('.');
                if (pos > 0) {
                    Ext = FileName.Mid(pos);
                    Ext.MakeLower();
                    if (Ext == ".cr2" || Ext == ".jpg") { 
                        CString     SrcFile, DstFile;

                        SrcFile = Dir + "\\" + FileName;

                        if (g_DstDir == "..") {

                            DstFile = Dir.Left(Dir.ReverseFind('\\')+1) + FileName;

                        } else {

                            // Strip off SrcDir and pre-pend DstDir
                            DstFile = g_DstDir + SrcFile.Mid(g_SrcDir.GetLength());
                        }

                        pos = DstFile.ReverseFind('.');
                        DstFile = DstFile.Left(pos) + ".jpg";
                        ProcessFile(SrcFile, DstFile);
                    }
                }
            }
        } while (FindNextFile(h, &data));
    }
    FindClose(h);

    //
    // Walk the sub-dirs
    //

    if (g_Recurs) {
        FileName = Dir + "\\*";
        h = FindFirstFile(FileName, &data);
        if (h != INVALID_HANDLE_VALUE) {
            do {
                if (data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
                    FileName = data.cFileName;
                    if (FileName != "." && FileName != "..") {
                        FileName = Dir + "\\" + FileName;
                        WalkDir(FileName);
                    }
                }
            } while (FindNextFile(h, &data));
        }
        FindClose(h);
    }
}


VOID ProcessFile(CString SrcFile, CString DstFile)
{
    CExifFile           Src;
    CExifFile           Dst;

    g_FileCount += 1;

    // if src == dst or no dest file, do nothing
    if (SrcFile == DstFile || !IsFile(DstFile)) {
        return ;
    }

    // parse exif info in files
    try {
      
        // Open & parse files
        Dst.Open(DstFile);
        Src.Open(SrcFile);
        g_CheckCount += 1;

        // Diff for updates
        Dst.CheckExifInfo(Src);

        // Apply updates to target
        if (!g_DontUpdate) {
            Dst.ApplyUpdates();
        }

    } catch (CExifException *e) {
        printf("CopyExif: '%s' %s\n", e->m_FileName, e->m_Message);
        delete e;
    }
}




CExifFile::CExifFile()
{
    m_FileOpened     = FALSE;
    m_Valid          = FALSE;
    m_IfdOffset      = 0;
    m_IfdDirPos      = 0;
    m_Endian         = FALSE;
    m_NextTag        = 0;
    m_Diff           = 0;
    m_OutFileOpened  = 0;
    m_OutEOFPos      = 0;
    memset(&m_Hdr, 0, sizeof(m_Hdr));

    m_Main.m_File    = this;
    m_Main.m_Type    = "IFD";

    m_Sub.m_File     = this;
    m_Sub.m_Type     = "SUB";

    m_Maker.m_File   = this;
    m_Maker.m_Type   = "MAK";
}


CExifFile::~CExifFile()
{
    Close();
}


VOID CExifFile::Close()
{
    if (m_FileOpened) {
        m_File.Close();
        m_FileOpened = FALSE;
    }

    if (m_OutFileOpened) {
        m_OutFile.Close();
        m_OutFileOpened = FALSE;
    }
}

VOID CExifFile::Open(CString FileName)
{
    INTN        pos;

    ASSERT(m_FileName.GetLength() == 0);
    ASSERT(!m_FileOpened);

    m_FileName = FileName;
    pos = m_FileName.ReverseFind('.');
    if (pos >= 0) {
        m_FileExt  = m_FileName.Mid(pos);
    }

    try {
      
        // Open & parse source file exif info
        IOpen();
        m_Valid = TRUE;

    } catch (CExifException *e) {
        e->m_FileName = m_FileName;
        throw ;
    }
}


VOID CExifFile::IOpen()
{
    UINT16          u16;
    CExifItem       *Item;
    
    m_FileOpened = m_File.Open(m_FileName, CFile::modeRead | CFile::shareDenyWrite | CFile::typeBinary);
    if (!m_FileOpened) {
        throw new CExifException("Failed to open");
    }

    //
    // Determine header type
    //

    u16 = ReadU16();
    if (u16 == 0x4949) {            // Canon CR2 file
        m_NextTag = 0xFFFF;
        ReadSeek(0x10);             // BUGBUG: hardcoded

    } else if (u16 == 0xd8ff) {     // JPG file
        ReadSeek(0);
        Read(&m_Hdr, sizeof(m_Hdr));

        // for now - onyl little endian format
        if (m_Hdr.SOI != 0xd8ff || m_Hdr.App1 != 0xe1ff || m_Hdr.Type != 0x4949) {
            throw new CExifException ("Jpg header type not supported");
        }

        m_NextTag = m_Hdr.Size;
        Swiz16((UCHAR *) &m_NextTag);
        m_NextTag += offsetof(EXIF_HEADER, Size);

        m_IfdOffset = 0xC;
        ReadSeek(8);

    } else {
        throw new CExifException ("File not support (header not understood)");
    }


    //
    // Parse IFD0, IFD1, ..., SUB IFD, and Maker Note
    //

    DbgPrint(VERBOSE_PARSE, "%s parse:\n", m_FileName);
    ReadIFD(&m_Main);
    Item = m_Main.Get(0x8769);              // Get sub tag offset
    if (Item) {
        Item->m_DontCheck = TRUE;
        ReadSeek(Item->GetU32());
        ReadIFD(&m_Sub);
    }

    Item = m_Sub.Get(0x927c);               // Get maker note
    if (Item) {
        Item->m_DontCheck = TRUE;
        Item->m_Size = 4;                   // dumby data
        ReadSeek(Item->m_ReadDataPos);
        ReadIFD(&m_Maker);
    }

    
    //
    // If there's a thumbnail location in IFD1 grab it
    //

    if (m_Main.m_NextIFD) {
        CExifItem   *ILen;

        ILen = m_Main.m_NextIFD->Get(0x202);
        Item = m_Main.m_NextIFD->Get(0x201);

        if (ILen && Item) {
            ReadSeek(Item->GetU32());
            Item->m_Size = ILen->GetU32();
            free(Item->m_Buffer);
            Item->m_Buffer = (UCHAR *) malloc(Item->m_Size);
            Read(Item->m_Buffer, Item->m_Size);
        }
    }

    DbgPrint(VERBOSE_PARSE, "\n");
}


VOID CExifFile::ReadIFD(CExifList *List)
{
    UINTN       NoDir;
    UINT64      DirPos;
    UINTN       DirDbgPos;
    UINT16      TagNo, Format;
    UINTN       NoComp, Size;
    CExifItem   *Item;
    CHAR        DataPosStr[32];

    NoDir  = ReadU16();
    DirPos = ReadPos();

    DbgPrint(VERBOSE_PARSE, "  %s%d %d entries starting at offset 0x%04x\n", List->m_Type, List->m_Level, NoDir, ReadDbgPos()-2);

    while (NoDir) {
        ReadSeek(DirPos);
        DirDbgPos = ReadDbgPos();

        TagNo  = ReadU16();
        Format = ReadU16();
        NoComp = ReadU32();

        if (Format > arrsize(g_ExifFormat)) {
            throw new CExifException ("Unknown format type");
        }

        //
        // Get base size of component elements
        //

        Size = g_ExifFormat[Format].Size;
        switch(TagNo) {     // Handle special cases where format is not supplied with the tag
        case 0x9C9B:
        case 0x9C9C:
        case 0x9C9D:
        case 0x9C9E:
        case 0x9C9F:
            Size = 2;        // unicode string
            break;
        }

        //
        // Get size of Exif data item
        //

        Size = Size * NoComp;
        if (Size > 64*1024) {
            throw new CExifException ("Parse error - size too big");
        }

        //
        // Allocate & read the contents of this tag
        //

        Item = new CExifItem(List, TagNo, Format, NoComp, Size);
        List->m_Tags.LinkTail(Item);

        strcpy(DataPosStr, "------");
        if (Size > 4) {
            Item->m_ReadDataPos = ReadU32();
            ReadSeek(Item->m_ReadDataPos);
            sprintf(DataPosStr, "0x%04x", ReadDbgPos());
        }

        Read(Item->m_Buffer, Size);

        DbgPrint(VERBOSE_PARSE, "    Dir Entry 0x%04x, Tag %s%d:0x%04x, Data %s, Size 0x%04x\n",
            DirDbgPos,
            List->m_Type, 
            List->m_Level,
            TagNo,
            DataPosStr,
            Size
            );

        //
        // Next item in Exif directory
        //

        NoDir -= 1;
        DirPos += 0xC;
    }

    // Read table terminator
    DirPos = ReadU32();

    // Canon doesn't put a terminitator on their Exif in CR2 files, or
    // in their maker note
    if (m_FileExt == ".cr2" || List->m_Type == "MAK") {
        DirPos = 0;
    }

    if (DirPos) {
        // Got another IFD, read it
        ASSERT(!List->m_NextIFD);
        List->m_NextIFD = new CExifList(List);

        ReadSeek(DirPos);
        ReadIFD(List->m_NextIFD);
    }
}

VOID CExifFile::CheckExifInfo(CExifFile &SrcFile)
{

    // Remove thumbnail?
    if (g_RemoveThumb && m_Main.m_NextIFD) {
        // IFD1 holds the embedded thumb
        m_Main.m_NextIFD->RemoveAll();
    }

    m_Main.Check(SrcFile.m_Main, this);
    m_Sub.Check(SrcFile.m_Sub, this);
    m_Maker.Check(SrcFile.m_Maker, this);

    if (m_Diff) {
        CExifItem       *Item;

        // Canon Digital Photo Professional refuses to display the Canon Maker tag
        // if this item is in the header.   Interestingly enough this app adds this
        // item when updating an image (and it removes the maker tag) !?
        Item = m_Main.Get(0x0131);
        if (Item && !Item->m_Remove) {
            Item->m_Remove = TRUE;
            m_Diff += 1;
        }

        DbgPrint(VERBOSE_DIFF, "%s diffs:\n", m_FileName);

        m_Main.DumpDiffs();
        m_Sub.DumpDiffs();
        m_Maker.DumpDiffs();

        DbgPrint(VERBOSE_DIFF, "\n");
    }
}

UINT64 CExifFile::ReadPos()
{
    return m_File.GetPosition() - m_IfdOffset;
}

UINTN CExifFile::ReadDbgPos()
{
    return (UINTN) m_File.GetPosition();
}

VOID  CExifFile::ReadSeek(UINT64 Pos)
{
    Pos += m_IfdOffset;
    if (Pos > m_NextTag) {
        throw new CExifException("Attempted to seek past end of section");
    }

    m_File.Seek(Pos, CFile::begin);
}

VOID  CExifFile::Read(VOID *Buffer, UINTN Size)
{
    m_File.Read(Buffer, Size);
}

UINT32 CExifFile::ReadU32()
{
    UINT32      u32;

    Read(&u32, sizeof(u32));
    return u32;
}

UINT16 CExifFile::ReadU16()
{
    UINT16      u16;

    Read(&u16, sizeof(u16));
    return u16;
}



VOID CExifFile::ApplyUpdates()
{
    CString         NewName;
    CString         OldName;
    UINT32          u32;
    UINT64          Len;
    CExifItem       *Sub = NULL;
    CExifItem       *Maker = NULL;

    if (!m_Diff) {
        printf("%s - no diffs\n", m_FileName);
        return ;
    }

    m_Main.Compute();
    m_Sub.Compute();
    m_Maker.Compute();

    printf("%s - %d diffs\n", m_FileName, m_Diff);
    g_UpdateCount += 1;

    NewName = m_FileName + ".tmp";
    OldName = m_FileName + ".old";

    m_OutFileOpened = m_OutFile.Open(NewName, CFile::modeCreate | CFile::modeReadWrite | CFile::typeBinary);
    if (!m_OutFileOpened) {
        throw new CExifException("Failed to open tmp file");
    }

    // 
    // Write header
    //

    Write(&m_Hdr, sizeof(m_Hdr));

    //
    // Write IFD directories
    //

    Sub = m_Main.Get(0x8769); 
    if (!Sub) {
        throw new CExifException("not done - target needs new sub ifd in ifd0");
    }

    m_Main.WriteDir();
    m_Sub.WriteDir();
    m_Maker.WriteDir();

    WriteU16(0xD9FF);
    m_OutEOFPos = WriteDbgPos();

    // 
    DbgPrint(VERBOSE_SAVE, "  IFD-EOF: 0x%04x.   Re-write:\n", m_OutEOFPos);

    Sub->m_Dispos = "SubPtr";
    memcpy(Sub->m_Buffer, &m_Sub.m_WriteDirHdrPos, sizeof(UINT32));
    Sub->RewriteEntry();
    
    Maker = m_Sub.Get(0x927c);
    if (Maker) {
        Maker->m_Dispos     = "Maker Note";
        Maker->m_Size       = (UINTN) (m_Maker.m_WriteDirEofPos - m_Maker.m_WriteDirHdrPos);
        Maker->m_BufferPos  = (UINTN) m_Maker.m_WriteDirHdrPos;
        Maker->m_NoComp     = Maker->m_Size;
        Maker->RewriteEntry();
    }

    //
    // Fix header
    //

    if (m_OutEOFPos > 0xFFF0) {
        throw new CExifException("Exif data too large");
    }

    m_IfdOffset = 0;
    WriteSeek(0);
    m_Hdr.Size = ((UINT16) m_OutEOFPos) - offsetof(EXIF_HEADER, Size);
    Swiz16((UCHAR * ) &m_Hdr.Size);
    Write(&m_Hdr, sizeof(m_Hdr));

    //
    // Copy the remaining part of the JPG file
    //

    m_OutFile.Seek(m_OutEOFPos, CFile::begin);
    m_File.Seek(m_NextTag, CFile::begin);
    Len = m_File.GetLength() - m_NextTag;
    DbgPrint(VERBOSE_SAVE, "  Append 0x%x bytes from offset 0x%x of original jpg\n", ((UINTN) Len), ((UINTN) m_NextTag));

    while (Len) {
        u32 = Len > G_BUFFER_SIZE ? G_BUFFER_SIZE : ((UINT32) Len);
        Read(g_Buffer, u32);
        Write(g_Buffer, u32);
        Len -= u32;
    }

    // Done
    m_OutFile.Close();
    m_OutFileOpened = FALSE;

    m_File.Close();
    m_FileOpened = FALSE;

    MoveFile(m_FileName, OldName);
    MoveFile(NewName, m_FileName);

    if (!g_DontDelete) {
        DeleteFile(OldName);
    }
}




UINT64 CExifFile::WritePos()
{
    return m_OutFile.GetPosition() - m_IfdOffset;
}

UINTN CExifFile::WriteDbgPos()
{
    return (UINTN) m_OutFile.GetPosition();
}

VOID  CExifFile::WriteSeek(UINT64 Pos)
{
    Pos += m_IfdOffset;
    m_OutFile.Seek(Pos, CFile::begin);
}


VOID CExifFile::Write(VOID *Data, UINTN Size)
{
    m_OutFile.Write(Data, Size);
}

VOID CExifFile::WriteU32(UINT32 u32)
{
    Write(&u32, sizeof(u32));
}

VOID CExifFile::WriteU16(UINT16 u16)
{
    Write(&u16, sizeof(u16));
}

VOID CExifFile::WriteU8(UINT8 u8)
{
    Write(&u8, sizeof(u8));
}

VOID CExifList::WriteDir()
{
    CExifItem       *Item;
    UINT64          DirPos, DataPos;

    DbgPrint(VERBOSE_SAVE, "  %s%d %d entries start at 0x%04x\n", m_Type, m_Level, m_Count, m_File->WriteDbgPos());

    // Start with entry count in this IFD
    m_WriteDirHdrPos = m_File->WritePos();
    m_File->WriteU16((UINT16) m_Count);
    DirPos  = m_File->WritePos();
    DataPos = DirPos + m_Count * 0xC + 4;

    // Write the entries
    for(Item=m_Tags.Head(); Item; Item=m_Tags.Next(Item)) {
        if (!Item->m_Remove) {
            DataPos += Item->WriteEntry(DirPos, DataPos);
            DirPos += 0xC;
        } else {
            DbgPrint(
                VERBOSE_SAVE, "    Dir Entry         Tag %s%d:0x%04x, Data         Size 0x%04x  remove\n",
                m_Type,
                m_Level,
                Item->m_TagNo,
                Item->m_Size
                );
        }
    }
    
    // tail
    m_File->WriteU32(m_NextIFD ? (UINTN) DataPos : 0 );
    m_File->WriteSeek(DataPos);
    m_WriteDirEofPos = DataPos;

    if (m_NextIFD) {
        m_NextIFD->WriteDir();
    }
}

VOID CExifItem::RewriteEntry()
{
    WriteEntry(m_WriteDirPos, m_WriteDataPos);
}


UINTN CExifItem::WriteEntry(UINT64 DirPos, UINT64 DataPos)
{
    CExifFile       *File;
    UINTN           DataUsed;
    CHAR            WriteDataPosStr[32];

    File = m_List->m_File;
    m_WriteDirPos = DirPos;

    File->WriteSeek(DirPos);
    File->WriteU16(m_TagNo);
    File->WriteU16(m_Format);
    File->WriteU32(m_NoComp);

    if (m_BufferPos != 0) {
        m_WriteDataPos = m_BufferPos;
        File->WriteU32(m_BufferPos);
        DataUsed = 0;

    } else  if (m_Size <= 4) {
        m_WriteDataPos = File->WritePos();
        File->Write(m_Buffer, m_Size);
        DataUsed = 0;

    } else {
        m_WriteDataPos = DataPos;
        File->WriteU32((UINT32) DataPos);
        File->WriteSeek(DataPos);
        File->Write(m_Buffer, m_Size);
        DataUsed = m_Size;
        DataUsed = (DataUsed + 1) & (~1);       // round

    }

    strcpy(WriteDataPosStr, "------");
    if (m_Size > 4) {
        sprintf(WriteDataPosStr, "0x%04x", (UINTN) m_WriteDataPos + File->m_IfdOffset);
    }

    DbgPrint(
        VERBOSE_SAVE, "    Dir Entry 0x%04x, Tag %s%d:0x%04x, Data %s, Size 0x%04x  %s\n",
        (UINTN) m_WriteDirPos + File->m_IfdOffset,
        m_List->m_Type,
        m_List->m_Level,
        m_TagNo,
        WriteDataPosStr,
        m_Size,
        m_Dispos
        );

    return DataUsed;
}



/////////////////////////////////////////////////////////////////////////////////

CExifItem::CExifItem(CExifList *Parent, UINT16 TagNo, UINT16 Format, UINT16 NoComp, UINTN Size) :
    m_List(Parent),
    m_TagNo(TagNo),
    m_Format(Format),
    m_NoComp(NoComp),
    m_Size(Size),
    m_ReadDataPos(0),
    m_WriteDirPos(0),
    m_WriteDataPos(0),
    m_BufferPos(0),
    m_Remove(FALSE),
    m_DontCheck(FALSE)
{
    m_Buffer = (UCHAR *) malloc(Size);
    ASSERT(m_Buffer);
}

CExifItem::~CExifItem()
{
    free(m_Buffer);
    m_Buffer = NULL;
}

UINT32 CExifItem::GetU32()
{
    UINT32      u32;

    ASSERT(m_Size == sizeof(u32));
    memcpy (&u32, m_Buffer, sizeof(u32));
    return u32;
}


/////////////////////////////////////////////////////////////////////////////////

CExifList::CExifList() :
    m_File(NULL),
    m_Level(0),
    m_NextIFD(NULL),
    m_Count(0),
    m_DirSize(0),
    m_DataSize(0),
    m_WriteDirHdrPos(0)
{
}

CExifList::CExifList(CExifList *Parent) :
    m_NextIFD(NULL),
    m_Count(0),
    m_DirSize(0),
    m_DataSize(0),
    m_WriteDirHdrPos(0)
{
    m_File  = Parent->m_File;
    m_Type  = Parent->m_Type;
    m_Level = Parent->m_Level+1;
}


CExifList::~CExifList()
{
    m_Tags.DeleteAll();
    SafeDelete(m_NextIFD);
}

CExifItem *CExifList::Get(UINT16 TagNo)
{
    CExifItem   *Item;

    for (Item=m_Tags.Head(); Item; Item=m_Tags.Next(Item)) {
        if (Item->m_TagNo == TagNo) {
            break;
        }
    }

    return Item;
}

VOID CExifList::Check(CExifList &SrcList, CExifFile *File)
{
    CExifItem   *SrcItem, *DstItem, *Sort;

    // Check each item to see if it's present in dest items
    for(SrcItem=SrcList.m_Tags.Head(); SrcItem; SrcItem=SrcList.m_Tags.Next(SrcItem)) {

        DstItem = Get(SrcItem->m_TagNo);
        if (DstItem) {

            // Item exists, does it need to match?
            if (g_Update && !SrcItem->m_DontCheck && !DstItem->m_DontCheck && !DstItem->m_Remove && 
                (DstItem->m_Size != SrcItem->m_Size || memcmp(DstItem->m_Buffer, SrcItem->m_Buffer, DstItem->m_Size) != 0)) {
                // replace
                free(DstItem->m_Buffer);
                DstItem->m_ReadDataPos = SrcItem->m_ReadDataPos;
                DstItem->m_Buffer = (UCHAR *) malloc(SrcItem->m_Size);
                DstItem->m_Size = SrcItem->m_Size;
                memcpy(DstItem->m_Buffer, SrcItem->m_Buffer, DstItem->m_Size);
                DstItem->m_Dispos = "change";
            }

        } else {

            // Item is missing - add it
            DstItem = new CExifItem(this, SrcItem->m_TagNo, SrcItem->m_Format, SrcItem->m_NoComp, SrcItem->m_Size);

            for(Sort=m_Tags.Head(); Sort; Sort=m_Tags.Next(Sort)) {
                if (Sort->m_TagNo > SrcItem->m_TagNo) {
                    break;
                }
            }

            if (Sort) {
                Sort->m_Link.LinkBefore(&DstItem->m_Link);
            } else {
                m_Tags.LinkTail(DstItem);
            }

            memcpy(DstItem->m_Buffer, SrcItem->m_Buffer, DstItem->m_Size);
            DstItem->m_ReadDataPos = SrcItem->m_ReadDataPos;
            DstItem->m_Dispos = "add";
        }
        
        if (DstItem && DstItem->m_Dispos.GetLength()) {
            m_File->m_Diff += 1;
        }
    }

    if (g_Replace) {
        for (DstItem = m_Tags.Head(); DstItem; DstItem = m_Tags.Next(DstItem)) {
            if (!SrcList.Get(DstItem->m_TagNo) && !DstItem->m_Remove) {
                DstItem->m_Remove = TRUE;
                m_File->m_Diff += 1;
            }
        }
    }

}

VOID CExifList::RemoveAll()
{
    CExifItem       *Item;
    
    for(Item=m_Tags.Head(); Item; Item=m_Tags.Next(Item)) {
        if (!Item->m_Remove) {
            Item->m_Remove += 1;
            m_File->m_Diff += 1;
        }
    }
}

VOID CExifList::DumpDiffs()
{
    CExifItem       *Item;
    UINTN           Count;

    Count = 0;
    for(Item=m_Tags.Head(); Item; Item=m_Tags.Next(Item)) {
        if (Item->m_Remove) {
            Item->m_Dispos = "remove";
        }

        if (Item->m_Dispos.GetLength()) {
            Count += 1;
            DbgPrint(
                VERBOSE_DIFF, "    %02d %s%d:0x%04x %-8s - %d bytes\n", 
                Count,
                m_Type, 
                m_Level, 
                Item->m_TagNo, 
                Item->m_Dispos,
                Item->m_Size
                );
        }
    }
}


VOID CExifList::Compute()
{
    CExifItem   *Item;

    m_Count = 0;
    m_DirSize = 0;
    m_DataSize = 0;

    if (m_NextIFD) {
        m_NextIFD->Compute();
    }

    m_DirSize += 2;     // # of entries at start of dir
    for(Item=m_Tags.Head(); Item; Item=m_Tags.Next(Item)) {
        if (!Item->m_Remove) {
            m_Count     += 1;
            m_DirSize   += 0xC;
            m_DataSize  += Item->m_Size > 4 ? Item->m_Size : 0;
            m_DataSize   = (m_DataSize + 1) & (~1);     // round
        }
    }
    m_DirSize += 4;     // next IDF pointer at end of dir
}

