#pragma once

#include "resource.h"


class CExifItem;
class CExifList;
class CExifFile;


//////////////////////////////////////////////////////////////////////
//
//

extern UINTN    g_Verbose;


/////////////////////////////////////////////////////////////////////////////
//
// EXIF - file format structures
//

typedef struct {
    UINT16          SOI;        // 0xFFD8   - JPG file
    UINT16          App1;       // 0xFFE1   - Exif marker
    UINT16          Size;       // Size of App1 marker
    UCHAR           Exif[6];    // Exif\0\0 
    UINT16          Type;       // II or MM
    UINT16          TagMark;    // 0x2a
    UINT32          Unkn;       // what is this?
} EXIF_HEADER;


// Note the IFD is preceeded by a 2 byte value that is the # of entries
// and followed by a 4 byte value that is the offset to the next IFD. 

typedef struct {
    UINT16          Tag;        // The Tag # for this entry
    UINT16          Format;     // Format of the data
    UINT32          NoComp;     // Number of compents

    union {
        UCHAR       Data[4];
        UINT32      Offset;
    } u;
} IFD_ENTRY ;

typedef struct {
    UINTN       Size;                   // Size of each item in this format
    VOID  (*Swizzle)(UCHAR *Data);      // Swizzle the data
} EXIF_FORMAT;


//////////////////////////////////////////////////////////////////////
//
//

class CExifItem {
public:
    CExifItem(CExifList *Parent, UINT16 TagNo, UINT16 Format, UINT16 NoComp, UINTN Size);
    ~CExifItem();

    UINT32  GetU32();
    UINTN   WriteEntry(UINT64 DirPos, UINT64 DataPos);
    VOID    RewriteEntry();

public:
    CExifList               *m_List;
    CLink                   m_Link;
    CString                 m_Dispos;

    UINT16                  m_TagNo;
    UINT16                  m_Format;
    UINT16                  m_NoComp;

    UINT32                  m_ReadDataPos;      // Position read in at
    UINTN                   m_Size;
    UCHAR                   *m_Buffer;
    UINTN                   m_BufferPos;        // If at a well-defined location

    BOOL                    m_Remove;
    BOOL                    m_DontCheck;

    // If written
    UINT64                  m_WriteDirPos;    
    UINT64                  m_WriteDataPos;
};


class CExifList {
public:
    CExifList();
    CExifList(CExifList *Parent);
    ~CExifList();
    CExifItem   *Get(UINT16 TagNo);

    VOID Check(CExifList &Src, CExifFile *File);
    VOID RemoveAll();
    VOID DumpDiffs();
    VOID Compute();
    VOID WriteDir();

public:
    CExifFile                   *m_File;
    CString                     m_Type;
    UINTN                       m_Level;

    TLIST(CExifItem, m_Link)    m_Tags;         // The tags
    CExifList                   *m_NextIFD;     

    // After Compute
    UINTN                       m_Count;
    UINTN                       m_DirSize;
    UINTN                       m_DataSize;

    // If written
    UINT64                      m_WriteDirHdrPos; 
    UINT64                      m_WriteDirEofPos; 
};


class CExifFile {
public:
    CExifFile();
    ~CExifFile();

    VOID Open(CString FileName);
    VOID Close();

    VOID CheckExifInfo(CExifFile &Src);
    VOID ApplyUpdates();

private:
    VOID IOpen();

public:
    CString                     m_FileName;
    CString                     m_FileExt;
    BOOL                        m_Valid;
    CExifList                   m_Main;         // The main tag list
    CExifList                   m_Sub;          // The sub tag list
    CExifList                   m_Maker;        

    // List updates that occured
    UINTN                       m_Diff;         // Items differ by

private:
    // Parsing
    CStdioFile                  m_File;
    BOOL                        m_FileOpened;
    BOOL                        m_Endian;
    EXIF_HEADER                 m_Hdr;
    UINTN                       m_IfdOffset;
    UINTN                       m_IfdDirPos;
    UINTN                       m_NextTag;      // Tag after Exif info

    VOID    ReadSeek(UINT64 Pos);
    UINT64  ReadPos();
    UINTN   ReadDbgPos();
    UINT16  ReadU16();
    UINT32  ReadU32();
    VOID    Read(VOID *Buffer, UINTN Size);
    VOID    ReadIFD(CExifList *List);


    // Output parsing
    CStdioFile                  m_OutFile;
    BOOL                        m_OutFileOpened;
    UINT64                      m_OutEOFPos;

    VOID    WriteSeek(UINT64 Pos);
    UINT64  WritePos();
    UINTN   WriteDbgPos();
    UINT64  WriteData(UCHAR *Buffer, UINTN Size);
    VOID    Write(VOID *Buffer, UINTN Size);
    VOID    WriteU32(UINT32 u32);
    VOID    WriteU16(UINT16 u16);
    VOID    WriteU8(UINT8 u8);

friend class CExifList;
friend class CExifItem;
};



class CExifException {
public:
    CExifException(LPCSTR Msg, ...);

public:
    CString                 m_FileName;
    CString                 m_Message;
} ; 

