/****************************************************************************
 *
 *  llist.h
 *  Double linked list
 *
 *
 *
 *  TLIST(N,L)
 *      N   - Type of objects in list
 *      L   - Link in Type
 *
 *  TADLIST(N,L)
 *      Same as TLIST expect that the list head destructor will
 *      automatically delete all nodes on the list
 *
 *
 *  By Ken Reneris              11/03/2003
 *
 ***/


class CLink {
private:
    CLink               *m_Prev;
    CLink               *m_Next;

//protected:
public:
    INLINE VOID LinkBefore (CLink *l);
    INLINE VOID LinkAfter (CLink *l);
    INLINE CLink *Next()                { return m_Next;                        }
    INLINE CLink *Prev()                { return m_Prev;                        }

public:
    INLINE bool IsLinked () const       { return m_Next != m_Prev;              }
    INLINE VOID Unlink ();

    INLINE CLink ()                     { m_Prev = this;  m_Next = this;        }
    INLINE ~CLink ()                    { ASSERT(!IsLinked());                  }
};


#define TLIST(N,L)    TList<N,offsetof(N,L),false>
#define TADLIST(N,L)  TList<N,offsetof(N,L),true>

template<class N, UINTN O, bool AD>
class TList
{
private:
    CLink               m_List;                 // List of entries
    UINTN               m_LinkOffset;           // Offset to CLink in class T

private:
    INLINE CLink *NTL(const N *n);              // Node-2-Link
    INLINE N *LTNC(const CLink *l);             // Link-2-Node with EOL check

public:
    INLINE N *Head ()                   { return LTNC(m_List.Next());           }
    INLINE N *Tail ()                   { return LTNC(m_List.Prev());           }
    INLINE N *Next (const N *n)         { return LTNC(NTL(n)->Next());          }
    INLINE N *Prev (const N *n)         { return LTNC(NTL(n)->Prev());          }

    INLINE VOID Unlink (N *n)           { NTL(n)->Unlink();                     }
    INLINE VOID LinkHead (N *n)         { m_List.LinkAfter(NTL(n));             }
    INLINE VOID LinkTail (N *n)         { m_List.LinkBefore(NTL(n));            }
    INLINE VOID Link (N *n)             { LinkTail(n);                          }

    VOID DeleteAll ();

    INLINE TList ()                     { m_LinkOffset = O;                     }
    ~TList ();
};



///////////////////////////////////////////////////////////////////////////
template<class N, UINTN O, bool AD>
TList<N,O,AD>::~TList ()
{
    if(AD) {
        DeleteAll();
    } else {
        ASSERT(!Head());
    }
}


template<class N, UINTN O, bool AD>
CLink *TList<N,O,AD>::NTL (const N *n)
{
    return (CLink *) (((UINT8 *) n) + m_LinkOffset);
}

template<class N, UINTN O, bool AD>
N *TList<N,O,AD>::LTNC (const CLink *l)
{
    return (N *) (l == &m_List ? NULL : ( ((UINT8 *) l) - m_LinkOffset ));
}

template<class N, UINTN O, bool AD>
VOID TList<N,O,AD>::DeleteAll ()
{
    N   *n;

    while ( (n = Head()) != NULL ) {
        Unlink(n);
        delete n;
    }
}


///////////////////////////////////////////////////////////////////////////
VOID CLink::Unlink ()
{
    m_Prev->m_Next = m_Next;
    m_Next->m_Prev = m_Prev;
    m_Next = this;
    m_Prev = this;
}

VOID CLink::LinkAfter (CLink *l)
{
    ASSERT(!l->IsLinked());
    l->m_Next = m_Next;
    l->m_Prev = this;
    m_Next->m_Prev = l;
    m_Next = l;

}

VOID CLink::LinkBefore (CLink *l)
{
    ASSERT(!l->IsLinked());
    l->m_Prev = m_Prev;
    l->m_Next = this;
    m_Prev->m_Next = l;
    m_Prev = l;
}
